@@ -1,35 +1,32 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
+from collections import defaultdict
+
 from sql import Null
+from sql.functions import CharLength
 
 from trytond.i18n import gettext
-from trytond.model import Workflow, Model, ModelView, ModelSQL, fields, Check
+from trytond.model import (
+    Check, Index, Model, ModelSQL, ModelView, Workflow, fields)
 from trytond.model.exceptions import AccessError
-from trytond.pyson import Eval, Bool, If
-from trytond.transaction import Transaction
 from trytond.pool import Pool
-from trytond.wizard import Wizard, StateView, StateTransition, Button
+from trytond.pyson import Bool, Eval, If
+from trytond.tools import grouped_slice, is_full_text, lstrip_wildcard
+from trytond.transaction import Transaction
+from trytond.wizard import Button, StateTransition, StateView, Wizard
 
-from .exceptions import InventoryValidationError, InventoryCountWarning
-
-__all__ = ['Inventory', 'InventoryLine',
-    'Count', 'CountSearch', 'CountQuantity']
-
-STATES = {
-    'readonly': Eval('state') != 'draft',
-}
-DEPENDS = ['state']
-INVENTORY_STATES = [
-    ('draft', 'Draft'),
-    ('done', 'Done'),
-    ('cancel', 'Canceled'),
-    ]
+from .exceptions import (
+    InventoryCountWarning, InventoryFutureWarning, InventoryValidationError)
 
 
 class Inventory(Workflow, ModelSQL, ModelView):
-    'Stock Inventory'
     __name__ = 'stock.inventory'
     _rec_name = 'number'
+
+    _states = {
+        'readonly': Eval('state') != 'draft',
+        }
+
     number = fields.Char('Number', readonly=True,
         help="The main identifier for the inventory.")
     location = fields.Many2One(
@@ -37,76 +34,77 @@ class Inventory(Workflow, ModelSQL, ModelView):
         domain=[('type', '=', 'storage')], states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
             },
-        depends=['state'],
         help="The location inventoried.")
     date = fields.Date('Date', required=True, states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
             },
-        depends=['state'],
         help="The date of the stock count.")
-    lost_found = fields.Many2One(
-        'stock.location', 'Lost and Found', required=True,
-        domain=[('type', '=', 'lost_found')], states=STATES, depends=DEPENDS,
-        help="Used for the balancing entries needed when the stock is "
-        "corrected.")
     lines = fields.One2Many(
         'stock.inventory.line', 'inventory', 'Lines',
         states={
-            'readonly': (STATES['readonly'] | ~Eval('location')
+            'readonly': (_states['readonly'] | ~Eval('location')
                 | ~Eval('date')),
-            },
-        depends=['location', 'date'] + DEPENDS)
+            })
     empty_quantity = fields.Selection([
             (None, ""),
             ('keep', "Keep"),
             ('empty', "Empty"),
-            ], "Empty Quantity", states=STATES, depends=DEPENDS,
-        help="How lines without quantity are handled.")
+            ], "Empty Quantity", states=_states,
+        help="How lines without a quantity are handled.")
     company = fields.Many2One('company.company', 'Company', required=True,
         states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
             },
-        depends=['state'],
         help="The company the inventory is associated with.")
-    state = fields.Selection(
-        INVENTORY_STATES, 'State', readonly=True, select=True,
+    state = fields.Selection([
+            ('draft', "Draft"),
+            ('done', "Done"),
+            ('cancelled', "Cancelled"),
+            ], "State", readonly=True, sort=False,
         help="The current state of the inventory.")
 
+    del _states
+
     @classmethod
     def __setup__(cls):
-        super(Inventory, cls).__setup__()
+        cls.number.search_unaccented = False
+        super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t,
+                (t.state, Index.Equality(cardinality='low')),
+                where=t.state == 'draft'))
         cls._order.insert(0, ('date', 'DESC'))
         cls._transitions |= set((
                 ('draft', 'done'),
-                ('draft', 'cancel'),
+                ('draft', 'cancelled'),
                 ))
         cls._buttons.update({
                 'confirm': {
-                    'invisible': Eval('state').in_(['done', 'cancel']),
+                    'invisible': Eval('state').in_(['done', 'cancelled']),
                     'depends': ['state'],
                     },
                 'cancel': {
-                    'invisible': Eval('state').in_(['cancel', 'done']),
+                    'invisible': Eval('state').in_(['cancelled', 'done']),
                     'depends': ['state'],
                     },
                 'complete_lines': {
                     'readonly': Eval('state') != 'draft',
                     'depends': ['state'],
                     },
-                'count': {
+                'do_count': {
                     'readonly': Eval('state') != 'draft',
                     'depends': ['state'],
                     },
                 })
 
     @classmethod
-    def __register__(cls, module_name):
-        super(Inventory, cls).__register__(module_name)
-
-        table = cls.__table_handler__(module_name)
-
-        # Add index on create_date
-        table.index_action('create_date', action='add')
+    def order_number(cls, tables):
+        table, _ = tables[None]
+        return [
+            ~((table.state == 'cancelled') & (table.number == Null)),
+            CharLength(table.number), table.number]
 
     @staticmethod
     def default_state():
@@ -121,35 +119,63 @@ class Inventory(Workflow, ModelSQL, ModelView):
     def default_company():
         return Transaction().context.get('company')
 
-    @classmethod
-    def default_lost_found(cls):
-        Location = Pool().get('stock.location')
-        locations = Location.search(cls.lost_found.domain)
-        if len(locations) == 1:
-            return locations[0].id
+    def get_rec_name(self, name):
+        pool = Pool()
+        Lang = pool.get('ir.lang')
+        lang = Lang.get()
+        date = lang.strftime(self.date)
+        return f"[{self.number}] {self.location.rec_name} @ {date}"
 
     @classmethod
-    def view_attributes(cls):
-        return [
-            ('/tree', 'visual', If(Eval('state') == 'cancel', 'muted', '')),
+    def search_rec_name(cls, name, clause):
+        _, operator, operand, *extra = clause
+        if operator.startswith('!') or operator.startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
+        number_value = operand
+        if operator.endswith('like') and is_full_text(operand):
+            number_value = lstrip_wildcard(operand)
+        return [bool_op,
+            ('number', operator, number_value, *extra),
+            ('location.rec_name', operator, operand, *extra),
             ]
 
     @classmethod
-    def delete(cls, inventories):
-        # Cancel before delete
-        cls.cancel(inventories)
-        for inventory in inventories:
-            if inventory.state != 'cancel':
-                raise AccessError(
-                    gettext('stock.msg_inventory_delete_cancel',
-                        inventory=inventory.rec_name))
-        super(Inventory, cls).delete(inventories)
+    def view_attributes(cls):
+        return super().view_attributes() + [
+            ('/tree', 'visual', If(Eval('state') == 'cancelled', 'muted', '')),
+            ]
 
     @classmethod
     @ModelView.button
     @Workflow.transition('done')
     def confirm(cls, inventories):
-        Move = Pool().get('stock.move')
+        pool = Pool()
+        Move = pool.get('stock.move')
+        Date = pool.get('ir.date')
+        Warning = pool.get('res.user.warning')
+        transaction = Transaction()
+        today_cache = {}
+
+        def in_future(inventory):
+            if inventory.company not in today_cache:
+                with Transaction().set_context(company=inventory.company.id):
+                    today_cache[inventory.company] = Date.today()
+            today = today_cache[inventory.company]
+            if inventory.date > today:
+                return inventory
+        future_inventories = sorted(filter(in_future, inventories))
+        if future_inventories:
+            names = ', '.join(i.rec_name for i in future_inventories[:5])
+            if len(future_inventories) > 5:
+                names + '...'
+            warning_name = Warning.format('date_future', future_inventories)
+            if Warning.check(warning_name):
+                raise InventoryFutureWarning(warning_name,
+                    gettext('stock.msg_inventory_date_in_the_future',
+                        inventories=names))
+
         moves = []
         for inventory in inventories:
             keys = set()
@@ -165,35 +191,49 @@ class Inventory(Workflow, ModelSQL, ModelView):
                 if move:
                     moves.append(move)
         if moves:
-            Move.save(moves)
-            Move.do(moves)
+            with transaction.set_context(_product_replacement=False):
+                Move.save(moves)
+            with transaction.set_context(_skip_warnings=True):
+                Move.do(moves)
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('cancel')
+    @Workflow.transition('cancelled')
     def cancel(cls, inventories):
         Line = Pool().get("stock.inventory.line")
         Line.cancel_move([l for i in inventories for l in i.lines])
 
     @classmethod
-    def create(cls, vlist):
+    def preprocess_values(cls, mode, values):
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Configuration = pool.get('stock.configuration')
-        config = Configuration(1)
-        vlist = [x.copy() for x in vlist]
-        for values in vlist:
-            if values.get('number') is None:
-                values['number'] = Sequence.get_id(
-                    config.inventory_sequence.id)
-        inventories = super(Inventory, cls).create(vlist)
-        cls.complete_lines(inventories, fill=False)
-        return inventories
+        values = super().preprocess_values(mode, values)
+        if mode == 'create' and not values.get('number'):
+            company_id = values.get('company', cls.default_company())
+            if company_id is not None:
+                configuration = Configuration(1)
+                if sequence := configuration.get_multivalue(
+                        'inventory_sequence', company=company_id):
+                    values['number'] = sequence.get()
+        return values
 
     @classmethod
-    def write(cls, inventories, values):
-        super(Inventory, cls).write(inventories, values)
-        cls.complete_lines(inventories, fill=False)
+    def on_modification(cls, mode, inventories, field_names=None):
+        super().on_modification(mode, inventories, field_names=field_names)
+        if mode in {'create', 'write'}:
+            cls.complete_lines(inventories, fill=False)
+
+    @classmethod
+    def check_modification(
+            cls, mode, inventories, values=None, external=False):
+        super().check_modification(
+            mode, inventories, values=values, external=external)
+        if mode == 'delete':
+            for inventory in inventories:
+                if inventory.state not in {'cancelled', 'draft'}:
+                    raise AccessError(gettext(
+                            'stock.msg_inventory_delete_cancel',
+                            inventory=inventory.rec_name))
 
     @classmethod
     def copy(cls, inventories, default=None):
@@ -208,9 +248,7 @@ class Inventory(Workflow, ModelSQL, ModelView):
         default.setdefault('lines.moves', None)
         default.setdefault('number', None)
 
-        new_inventories = super().copy(inventories, default=default)
-        cls.complete_lines(new_inventories, fill=False)
-        return new_inventories
+        return super().copy(inventories, default=default)
 
     @staticmethod
     def grouping():
@@ -227,33 +265,32 @@ class Inventory(Workflow, ModelSQL, ModelView):
         Product = pool.get('product.product')
 
         grouping = cls.grouping()
-        to_create, to_write = [], []
+        to_save, to_delete = [], []
         for inventory in inventories:
             # Once done computation is wrong because include created moves
             if inventory.state == 'done':
                 continue
             # Compute product quantities
-            if fill:
-                product_ids = None
-            else:
-                product_ids = [l.product.id for l in inventory.lines]
-            with Transaction().set_context(stock_date_end=inventory.date):
-                pbl = Product.products_by_location(
-                    [inventory.location.id],
-                    grouping=grouping,
-                    grouping_filter=(product_ids,))
-
-            # Index some data
-            product2type = {}
-            product2consumable = {}
-            for product in Product.browse({line[1] for line in pbl}):
-                product2type[product.id] = product.type
-                product2consumable[product.id] = product.consumable
+            with Transaction().set_context(
+                    company=inventory.company.id,
+                    stock_date_end=inventory.date):
+                if fill:
+                    pbl = Product.products_by_location(
+                        [inventory.location.id],
+                        grouping=grouping)
+                else:
+                    product_ids = [l.product.id for l in inventory.lines]
+                    pbl = defaultdict(int)
+                    for product_ids in grouped_slice(product_ids):
+                        pbl.update(Product.products_by_location(
+                                [inventory.location.id],
+                                grouping=grouping,
+                                grouping_filter=(list(product_ids),)))
 
             # Update existing lines
             for line in inventory.lines:
                 if line.product.type != 'goods':
-                    Line.delete([line])
+                    to_delete.append(line)
                     continue
 
                 key = (inventory.location.id,) + line.unique_key
@@ -261,62 +298,72 @@ class Inventory(Workflow, ModelSQL, ModelView):
                     quantity = pbl.pop(key)
                 else:
                     quantity = 0.0
-                values = line.update_values4complete(quantity)
-                if values:
-                    to_write.extend(([line], values))
+                line.update_for_complete(quantity)
+                to_save.append(line)
 
             if not fill:
                 continue
+
+            product_idx = grouping.index('product') + 1
+            # Index some data
+            product2type = {}
+            product2consumable = {}
+            for product in Product.browse({line[product_idx] for line in pbl}):
+                product2type[product.id] = product.type
+                product2consumable[product.id] = product.consumable
+
             # Create lines if needed
             for key, quantity in pbl.items():
-                product_id = key[grouping.index('product') + 1]
+                product_id = key[product_idx]
                 if (product2type[product_id] != 'goods'
                         or product2consumable[product_id]):
                     continue
                 if not quantity:
                     continue
 
-                values = Line.create_values4complete(inventory, quantity)
-                for i, fname in enumerate(grouping, 1):
-                    values[fname] = key[i]
-                to_create.append(values)
-        if to_create:
-            Line.create(to_create)
-        if to_write:
-            Line.write(*to_write)
+                line = Line(
+                    inventory=inventory,
+                    **{fname: key[i] for i, fname in enumerate(grouping, 1)})
+                line.update_for_complete(quantity)
+                to_save.append(line)
+        if to_delete:
+            Line.delete(to_delete)
+        if to_save:
+            Line.save(to_save)
 
     @classmethod
     @ModelView.button_action('stock.wizard_inventory_count')
-    def count(cls, inventories):
+    def do_count(cls, inventories):
         cls.complete_lines(inventories)
 
 
 class InventoryLine(ModelSQL, ModelView):
-    'Stock Inventory Line'
     __name__ = 'stock.inventory.line'
     _states = {
         'readonly': Eval('inventory_state') != 'draft',
         }
-    _depends = ['inventory_state']
 
     product = fields.Many2One('product.product', 'Product', required=True,
         domain=[
             ('type', '=', 'goods'),
-            ], states=_states, depends=_depends)
-    uom = fields.Function(fields.Many2One('product.uom', 'UOM',
-        help="The unit in which the quantity is specified."), 'get_uom')
-    unit_digits = fields.Function(fields.Integer('Unit Digits'),
-            'get_unit_digits')
-    expected_quantity = fields.Float('Expected Quantity', required=True,
-        digits=(16, Eval('unit_digits', 2)), readonly=True,
+            ], states=_states)
+    unit = fields.Function(fields.Many2One(
+            'product.uom', "Unit",
+            help="The unit in which the quantity is specified."),
+        'get_unit')
+    expected_quantity = fields.Float(
+        "Expected Quantity", digits='unit', required=True, readonly=True,
         states={
             'invisible': Eval('id', -1) < 0,
         },
-        depends=['unit_digits'],
         help="The quantity the system calculated should be in the location.")
-    quantity = fields.Float('Quantity',
-        digits=(16, Eval('unit_digits', 2)),
-        states=_states, depends=['unit_digits'] + _depends,
+    quantity = fields.Float(
+        "Actual Quantity", digits='unit', states=_states,
+        domain=[
+            If(Eval('quantity', None),
+                ('quantity', '>=', 0),
+                ()),
+            ],
         help="The actual quantity found in the location.")
     moves = fields.One2Many('stock.move', 'origin', 'Moves', readonly=True)
     inventory = fields.Many2One('stock.inventory', 'Inventory', required=True,
@@ -324,15 +371,24 @@ class InventoryLine(ModelSQL, ModelView):
         states={
             'readonly': _states['readonly'] & Bool(Eval('inventory')),
             },
-        depends=_depends,
         help="The inventory the line belongs to.")
+    inventory_location = fields.Function(
+        fields.Many2One('stock.location', "Location"),
+        'on_change_with_inventory_location',
+        searcher='search_inventory_location')
+    inventory_date = fields.Function(
+        fields.Date("Date"),
+        'on_change_with_inventory_date',
+        searcher='search_inventory_date')
     inventory_state = fields.Function(
-        fields.Selection(INVENTORY_STATES, 'Inventory State'),
+        fields.Selection('get_inventory_states', "Inventory State",
+        depends={'inventory'}),
         'on_change_with_inventory_state')
 
     @classmethod
     def __setup__(cls):
-        super(InventoryLine, cls).__setup__()
+        super().__setup__()
+        cls.__access__.add('inventory')
         t = cls.__table__()
         cls._sql_constraints += [
             ('check_line_qty_pos', Check(t, t.quantity >= 0),
@@ -340,46 +396,38 @@ class InventoryLine(ModelSQL, ModelView):
             ]
         cls._order.insert(0, ('product', 'ASC'))
 
-    @classmethod
-    def __register__(cls, module_name):
-        cursor = Transaction().connection.cursor()
-        pool = Pool()
-        Move = pool.get('stock.move')
-        sql_table = cls.__table__()
-        move_table = Move.__table__()
-
-        super(InventoryLine, cls).__register__(module_name)
-
-        table = cls.__table_handler__(module_name)
-
-        # Migration from 3.0: use Move origin
-        if table.column_exist('move'):
-            cursor.execute(*sql_table.select(sql_table.id, sql_table.move,
-                    where=sql_table.move != Null))
-            for line_id, move_id in cursor.fetchall():
-                cursor.execute(*move_table.update(
-                        columns=[move_table.origin],
-                        values=['%s,%s' % (cls.__name__, line_id)],
-                        where=move_table.id == move_id))
-            table.drop_column('move')
-
-        # Migration from 4.6: drop required on quantity
-        table.not_null_action('quantity', action='remove')
-
-    @staticmethod
-    def default_unit_digits():
-        return 2
-
     @staticmethod
     def default_expected_quantity():
         return 0.
 
     @fields.depends('product')
     def on_change_product(self):
-        self.unit_digits = 2
         if self.product:
-            self.uom = self.product.default_uom
-            self.unit_digits = self.product.default_uom.digits
+            self.unit = self.product.default_uom
+
+    @fields.depends('inventory', '_parent_inventory.location')
+    def on_change_with_inventory_location(self, name=None):
+        return self.inventory.location if self.inventory else None
+
+    @classmethod
+    def search_inventory_location(cls, name, clause):
+        nested = clause[0][len(name):]
+        return [('inventory.location' + nested, *clause[1:])]
+
+    @fields.depends('inventory', '_parent_inventory.date')
+    def on_change_with_inventory_date(self, name=None):
+        if self.inventory:
+            return self.inventory.date
+
+    @classmethod
+    def search_inventory_date(cls, name, clause):
+        return [('inventory.date',) + tuple(clause[1:])]
+
+    @classmethod
+    def get_inventory_states(cls):
+        pool = Pool()
+        Inventory = pool.get('stock.inventory')
+        return Inventory.fields_get(['state'])['state']['selection']
 
     @fields.depends('inventory', '_parent_inventory.state')
     def on_change_with_inventory_state(self, name=None):
@@ -394,11 +442,8 @@ class InventoryLine(ModelSQL, ModelView):
     def search_rec_name(cls, name, clause):
         return [('product.rec_name',) + tuple(clause[1:])]
 
-    def get_uom(self, name):
-        return self.product.default_uom.id
-
-    def get_unit_digits(self, name):
-        return self.product.default_uom.digits
+    def get_unit(self, name):
+        return self.product.default_uom
 
     @property
     def unique_key(self):
@@ -423,7 +468,6 @@ class InventoryLine(ModelSQL, ModelView):
         '''
         pool = Pool()
         Move = pool.get('stock.move')
-        Uom = pool.get('product.uom')
 
         qty = self.quantity
         if qty is None:
@@ -436,13 +480,16 @@ class InventoryLine(ModelSQL, ModelView):
             else:
                 qty = 0.0
 
-        delta_qty = Uom.compute_qty(self.uom,
-            self.expected_quantity - qty,
-            self.uom)
+        delta_qty = self.unit.round(self.expected_quantity - qty)
         if delta_qty == 0.0:
             return
         from_location = self.inventory.location
-        to_location = self.inventory.lost_found
+        to_location = self.inventory.location.lost_found_used
+        if not to_location:
+            raise InventoryValidationError(
+                gettext('stock.msg_inventory_location_missing_lost_found',
+                    inventory=self.inventory.rec_name,
+                    location=self.inventory.location.rec_name))
         if delta_qty < 0:
             (from_location, to_location, delta_qty) = \
                 (to_location, from_location, -delta_qty)
@@ -452,46 +499,31 @@ class InventoryLine(ModelSQL, ModelView):
             to_location=to_location,
             quantity=delta_qty,
             product=self.product,
-            uom=self.uom,
+            unit=self.unit,
             company=self.inventory.company,
             effective_date=self.inventory.date,
             origin=self,
             )
 
-    def update_values4complete(self, quantity):
-        '''
-        Return update values to complete inventory
-        '''
-        values = {}
-        # if nothing changed, no update
-        if self.expected_quantity == quantity:
-            return values
-        values['expected_quantity'] = quantity
-        return values
+    @fields.depends('expected_quantity')
+    def update_for_complete(self, quantity):
+        if self.expected_quantity != quantity:
+            self.expected_quantity = quantity
 
     @classmethod
-    def create_values4complete(cls, inventory, quantity):
-        '''
-        Return create values to complete inventory
-        '''
-        return {
-            'inventory': inventory.id,
-            'expected_quantity': quantity,
-        }
-
-    @classmethod
-    def delete(cls, lines):
-        for line in lines:
-            if line.inventory_state not in {'cancel', 'draft'}:
-                raise AccessError(
-                    gettext('stock.msg_inventory_line_delete_cancel',
-                        line=line.rec_name,
-                        inventory=line.inventory.rec_name))
-        super(InventoryLine, cls).delete(lines)
+    def check_modification(cls, mode, lines, values=None, external=False):
+        super().check_modification(
+            mode, lines, values=values, external=external)
+        if mode == 'delete':
+            for line in lines:
+                if line.inventory_state not in {'cancelled', 'draft'}:
+                    raise AccessError(gettext(
+                            'stock.msg_inventory_line_delete_cancel',
+                            line=line.rec_name,
+                            inventory=line.inventory.rec_name))
 
 
 class Count(Wizard):
-    "Stock Inventory Count"
     __name__ = 'stock.inventory.count'
     start_state = 'search'
 
@@ -511,29 +543,28 @@ class Count(Wizard):
 
     def default_quantity(self, fields):
         pool = Pool()
-        Inventory = pool.get('stock.inventory')
         InventoryLine = pool.get('stock.inventory.line')
         Warning = pool.get('res.user.warning')
-        context = Transaction().context
-        inventory = Inventory(context['active_id'])
         values = {}
-        lines = InventoryLine.search(self.get_line_domain(inventory), limit=1)
+        lines = InventoryLine.search(
+            self.get_line_domain(self.record), limit=1)
         if not lines:
             warning_name = '%s.%s.count_create' % (
-                inventory, self.search.search)
+                self.record, self.search.search)
             if Warning.check(warning_name):
                 raise InventoryCountWarning(warning_name,
                     gettext('stock.msg_inventory_count_create_line',
                         search=self.search.search.rec_name))
-            line, = InventoryLine.create([self.get_line_values(inventory)])
+            line = self.get_line()
+            line.update_for_complete(0)
+            line.save()
         else:
             line, = lines
         values['line'] = line.id
         values['product'] = line.product.id
-        values['uom'] = line.uom.id
-        values['unit_digits'] = line.unit_digits
-        if line.uom.rounding == 1:
-            values['quantity_added'] = 1
+        values['unit'] = line.unit.id
+        if line.unit.rounding == 1:
+            values['quantity'] = 1.
         return values
 
     def get_line_domain(self, inventory):
@@ -546,28 +577,28 @@ class Count(Wizard):
             domain.append(('product', '=', self.search.search.id))
         return domain
 
-    def get_line_values(self, inventory):
+    def get_line(self):
         pool = Pool()
         Product = pool.get('product.product')
         InventoryLine = pool.get('stock.inventory.line')
-        values = InventoryLine.create_values4complete(inventory, 0)
+
+        line = InventoryLine(inventory=self.record)
         if isinstance(self.search.search, Product):
-            values['product'] = self.search.search.id
-        return values
+            line.product = self.search.search
+        return line
 
     def transition_add(self):
-        if self.quantity.line and self.quantity.quantity_added:
+        if self.quantity.line and self.quantity.quantity:
             line = self.quantity.line
             if line.quantity:
-                line.quantity += self.quantity.quantity_added
+                line.quantity += self.quantity.quantity
             else:
-                line.quantity = self.quantity.quantity_added
+                line.quantity = self.quantity.quantity
             line.save()
         return 'search'
 
 
 class CountSearch(ModelView):
-    "Stock Inventory Count"
     __name__ = 'stock.inventory.count.search'
 
     search = fields.Reference(
@@ -575,55 +606,38 @@ class CountSearch(ModelView):
             ('product.product', "Product"),
             ],
         required=True,
-        domain=[If(Eval('search_model') == 'product.product',
-                [
-                    ('type', '=', 'goods'),
-                    ('consumable', '=', False),
-                    ],
-                [])],
-        depends=['search_model'],
+        domain={
+            'product.product': [
+                ('type', '=', 'goods'),
+                ('consumable', '=', False),
+                ],
+            },
         help="The item that's counted.")
-    search_model = fields.Function(fields.Selection(
-        'get_search_models', "Search Model"),
-        'on_change_with_search_model')
 
     @classmethod
     def default_search(cls):
         return 'product.product,-1'
 
-    @classmethod
-    def get_search_models(cls):
-        return cls.fields_get(['search'])['search']['selection']
-
-    @fields.depends('search')
-    def on_change_with_search_model(self, name=None):
-        if self.search:
-            return self.search.__name__
-
 
 class CountQuantity(ModelView):
-    "Stock Inventory Count"
     __name__ = 'stock.inventory.count.quantity'
 
     line = fields.Many2One(
         'stock.inventory.line', "Line", readonly=True, required=True)
     product = fields.Many2One('product.product', "Product", readonly=True)
-    uom = fields.Many2One('product.uom', "UOM", readonly=True,
+    unit = fields.Many2One(
+        'product.uom', "Unit", readonly=True,
         help="The unit in which the quantities are specified.")
     total_quantity = fields.Float(
-        "Total Quantity", digits=(16, Eval('unit_digits', 2)),
-        readonly=True, depends=['unit_digits'],
+        "Total Quantity", digits='unit', readonly=True,
         help="The total amount of the line counted so far.")
 
-    quantity_added = fields.Float(
-        "Added Quantity", digits=(16, Eval('unit_digits', 2)), required=True,
-        depends=['unit_digits'],
+    quantity = fields.Float(
+        "Quantity", digits='unit', required=True,
         help="The quantity to add to the existing count.")
 
-    unit_digits = fields.Integer("Unit Digits", readonly=True)
-
-    @fields.depends('quantity_added', 'line')
-    def on_change_quantity_added(self):
+    @fields.depends('quantity', 'line')
+    def on_change_quantity(self):
         if self.line:
             self.total_quantity = (
-                (self.line.quantity or 0) + (self.quantity_added or 0))
+                (self.line.quantity or 0) + (self.quantity or 0))
