@@ -4,46 +4,48 @@ import datetime
 import operator
 from decimal import Decimal
 
+from sql import Column
+
+from trytond.cache import Cache
 from trytond.i18n import gettext
 from trytond.model import (
-    ModelView, ModelSQL, MatchMixin, ValueMixin, DeactivableMixin, fields,
+    DeactivableMixin, Index, MatchMixin, Model, ModelSQL, ModelView, fields,
     sequence_ordered, tree)
-from trytond import backend
-from trytond.cache import Cache
-from trytond.pyson import Eval, If
-from trytond.transaction import Transaction
-from trytond.pool import Pool, PoolMeta
+from trytond.modules.product import price_digits, round_price
+from trytond.pool import Pool
+from trytond.pyson import Eval, If, TimeDelta
 from trytond.tools import grouped_slice
-from trytond.tools.multivalue import migrate_property
+from trytond.transaction import (
+    Transaction, inactive_records, without_check_access)
 
 from .exceptions import LocationValidationError
 
-__all__ = ['Location', 'Party', 'PartyLocation', 'ProductsByLocationsContext',
-    'LocationLeadTime']
 
-STATES = {
-    'readonly': ~Eval('active'),
-}
-DEPENDS = ['active']
+class WarehouseWasteLocation(ModelSQL):
+    __name__ = 'stock.location.waste'
+
+    warehouse = fields.Many2One(
+        'stock.location', "Warehouse", required=True, ondelete='CASCADE',
+        domain=[('type', '=', 'warehouse')])
+    location = fields.Many2One(
+        'stock.location', "Waste Location", required=True, ondelete='CASCADE',
+        domain=[('type', '=', 'lost_found')])
 
 
 class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
-    "Stock Location"
     __name__ = 'stock.location'
     _default_warehouse_cache = Cache('stock.location.default_warehouse',
         context=False)
 
-    name = fields.Char("Name", size=None, required=True, states=STATES,
-        depends=DEPENDS, translate=True)
-    code = fields.Char("Code", size=None, states=STATES, depends=DEPENDS,
-        select=True,
+    name = fields.Char("Name", size=None, required=True, translate=True)
+    code = fields.Char(
+        "Code",
         help="The internal identifier used for the location.")
-    address = fields.Many2One("party.address", "Address",
+    address = fields.Many2One(
+        'party.address', "Address",
         states={
             'invisible': Eval('type') != 'warehouse',
-            'readonly': ~Eval('active'),
-            },
-        depends=['type', 'active'])
+            })
     type = fields.Selection([
         ('supplier', 'Supplier'),
         ('customer', 'Customer'),
@@ -53,17 +55,27 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
         ('production', 'Production'),
         ('drop', 'Drop'),
         ('view', 'View'),
-        ], 'Location type', states=STATES, depends=DEPENDS)
+        ], "Type",
+        help_selection={
+            'supplier': "Used as the source of stock received from suppliers.",
+            'customer': "Used as the destination for stock sent to customers.",
+            'lost_found': "Used for damages, discrepancies and wastage.",
+            'warehouse': (
+                "Regroup storage locations under a logistics warehouse."),
+            'storage': "Used to physically store goods.",
+            'production': (
+                "Used as the destination of components and the source of "
+                "finished products."),
+            'drop': "Used during the drop shipping process.",
+            'view': "Group locations logically.",
+            })
     type_string = type.translated('type')
-    parent = fields.Many2One("stock.location", "Parent", select=True,
+    parent = fields.Many2One(
+        "stock.location", "Parent", ondelete='CASCADE',
         left="left", right="right",
-        states={
-            'invisible': Eval('type') == 'warehouse',
-            },
-        depends=['type'],
         help="Used to add structure above the location.")
-    left = fields.Integer('Left', required=True, select=True)
-    right = fields.Integer('Right', required=True, select=True)
+    left = fields.Integer('Left', required=True)
+    right = fields.Integer('Right', required=True)
     childs = fields.One2Many("stock.location", "parent", "Children",
         help="Used to add structure below the location.")
     flat_childs = fields.Boolean(
@@ -75,71 +87,122 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
     input_location = fields.Many2One(
         "stock.location", "Input", states={
             'invisible': Eval('type') != 'warehouse',
-            'readonly': ~Eval('active'),
             'required': Eval('type') == 'warehouse',
             },
         domain=[
-            ('type', '=', 'storage'),
             ['OR',
-                ('parent', 'child_of', [Eval('id')]),
+                ('type', '=', 'storage'),
+                ('id', '=', Eval('storage_location', -1)),
+                ],
+            ['OR',
+                ('parent', 'child_of', [Eval('id', -1)]),
                 ('parent', '=', None),
                 ],
             ],
-        depends=['type', 'active', 'id'],
         help="Where incoming stock is received.")
     output_location = fields.Many2One(
         "stock.location", "Output", states={
             'invisible': Eval('type') != 'warehouse',
-            'readonly': ~Eval('active'),
             'required': Eval('type') == 'warehouse',
         },
         domain=[
-            ('type', '=', 'storage'),
             ['OR',
-                ('parent', 'child_of', [Eval('id')]),
+                ('type', '=', 'storage'),
+                ('id', '=', Eval('storage_location', -1)),
+                ],
+            ['OR',
+                ('parent', 'child_of', [Eval('id', -1)]),
                 ('parent', '=', None)]],
-        depends=['type', 'active', 'id'],
         help="Where outgoing stock is sent from.")
     storage_location = fields.Many2One(
         "stock.location", "Storage", states={
             'invisible': Eval('type') != 'warehouse',
-            'readonly': ~Eval('active'),
             'required': Eval('type') == 'warehouse',
         },
         domain=[
             ('type', 'in', ['storage', 'view']),
             ['OR',
-                ('parent', 'child_of', [Eval('id')]),
+                ('parent', 'child_of', [Eval('id', -1)]),
                 ('parent', '=', None)]],
-        depends=['type', 'active', 'id'],
         help="The top level location where stock is stored.")
     picking_location = fields.Many2One(
         'stock.location', 'Picking', states={
             'invisible': Eval('type') != 'warehouse',
-            'readonly': ~Eval('active'),
             },
         domain=[
             ('type', '=', 'storage'),
             ('parent', 'child_of', [Eval('storage_location', -1)]),
             ],
-        depends=['type', 'active', 'storage_location'],
         help="Where stock is picked from.\n"
         "Leave empty to use the storage location.")
+    lost_found_location = fields.Many2One(
+        'stock.location', "Lost and Found",
+        states={
+            'invisible': Eval('type') != 'warehouse',
+            },
+        domain=[
+            ('type', '=', 'lost_found'),
+            ],
+        help="Used, by inventories, when correcting stock levels "
+        "in the warehouse.")
+    waste_locations = fields.Many2Many(
+        'stock.location.waste', 'warehouse', 'location', "Waste Locations",
+        states={
+            'invisible': Eval('type') != 'warehouse',
+            },
+        domain=[
+            ('type', '=', 'lost_found'),
+            ],
+        help="The locations used for waste products from the warehouse.")
+    waste_warehouses = fields.Many2Many(
+        'stock.location.waste', 'location', 'warehouse', "Waste Warehouses",
+        states={
+            'invisible': Eval('type') != 'lost_found',
+            },
+        domain=[
+            ('type', '=', 'warehouse'),
+            ],
+        help="The warehouses that use the location for waste products.")
+
+    allow_pickup = fields.Boolean(
+        "Allow Pickup",
+        states={
+            'invisible': (
+                (Eval('type') != 'warehouse')
+                & ~Eval('address')),
+            })
+
     quantity = fields.Function(
-        fields.Float('Quantity',
-        help="The amount of stock in the location."),
+        fields.Float(
+            "Quantity", digits='quantity_uom',
+            help="The amount of stock in the location."),
         'get_quantity', searcher='search_quantity')
     forecast_quantity = fields.Function(
-        fields.Float('Forecast Quantity',
-        help="The amount of stock expected to be in the location."),
+        fields.Float(
+            "Forecast Quantity", digits='quantity_uom',
+            help="The amount of stock expected to be in the location."),
         'get_quantity', searcher='search_quantity')
-    cost_value = fields.Function(fields.Numeric('Cost Value',
-        help="The value of the stock in the location."),
+    quantity_uom = fields.Function(fields.Many2One(
+            'product.uom', "Quantity UoM",
+            help="The Unit of Measure for the quantities."),
+        'get_quantity_uom')
+    cost_value = fields.Function(fields.Numeric(
+            "Cost Value", digits=price_digits,
+            help="The value of the stock in the location."),
         'get_cost_value')
 
     @classmethod
     def __setup__(cls):
-        super(Location, cls).__setup__()
+        cls.code.search_unaccented = False
+        super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(t, (t.code, Index.Similarity())),
+                Index(
+                    t,
+                    (t.left, Index.Range(cardinality='high')),
+                    (t.right, Index.Range(cardinality='high'))),
+                })
         cls._order.insert(0, ('name', 'ASC'))
 
         parent_domain = [
@@ -161,7 +224,6 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
                     ('type', 'in', childs_mapping[type_]), ()))
         cls.parent.domain = parent_domain
         cls.childs.domain = childs_domain
-        cls.childs.depends.extend(['flat_childs', 'type'])
 
     @classmethod
     def _parent_domain(cls):
@@ -174,7 +236,7 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             'lost_found': ['lost_found'],
             'view': ['warehouse', 'view', 'storage'],
             'storage': ['warehouse', 'view', 'storage'],
-            'warehouse': [''],
+            'warehouse': ['view'],
             }
 
     @classmethod
@@ -187,36 +249,31 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
         return childs_domain
 
     @classmethod
-    def __register__(cls, module_name):
-        super(Location, cls).__register__(module_name)
-
-        table = cls.__table_handler__(module_name)
-        table.index_action(['left', 'right'], 'add')
-
-    @classmethod
-    def validate(cls, locations):
-        super(Location, cls).validate(locations)
+    def validate_fields(cls, locations, field_names):
+        super().validate_fields(locations, field_names)
         inactives = []
         for location in locations:
-            location.check_type_for_moves()
-            if not location.active:
+            location.check_type_for_moves(field_names)
+            if 'active' in field_names and not location.active:
                 inactives.append(location)
         cls.check_inactive(inactives)
 
-    def check_type_for_moves(self):
+    def check_type_for_moves(self, field_names=None):
         """ Check locations with moves have types compatible with moves. """
+        pool = Pool()
+        Move = pool.get('stock.move')
+        if field_names and 'type' not in field_names:
+            return
         invalid_move_types = ['warehouse', 'view']
-        Move = Pool().get('stock.move')
         if self.type in invalid_move_types:
-            # Use root to compute for all companies
-            with Transaction().set_user(0):
-                moves = Move.search([
-                        ['OR',
-                            ('to_location', '=', self.id),
-                            ('from_location', '=', self.id),
-                            ],
-                        ('state', 'not in', ['staging', 'draft']),
-                        ])
+            moves = Move.search([
+                    ['OR',
+                        ('to_location', '=', self.id),
+                        ('from_location', '=', self.id),
+                        ],
+                    ('state', 'not in', ['staging', 'draft']),
+                    ],
+                order=[], limit=1)
             if moves:
                 raise LocationValidationError(
                     gettext('stock.msg_location_invalid_type_for_moves',
@@ -238,27 +295,28 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
     def get_empty_locations(cls, locations=None):
         pool = Pool()
         Move = pool.get('stock.move')
+        Product = pool.get('product.product')
         if locations is None:
             locations = cls.search([])
         if not locations:
             return []
         location_ids = list(map(int, locations))
-        # Use root to compute for all companies
-        # and ensures inactive locations are in the query
-        with Transaction().set_user(0), \
-                Transaction().set_context(active_test=False):
+        with without_check_access(), inactive_records():
             query = Move.compute_quantities_query(
                 location_ids, with_childs=True)
             quantities = Move.compute_quantities(
                 query, location_ids, with_childs=True)
             empty = set(location_ids)
+            product_ids = [q[1] for q in quantities.keys()]
+            consumables = {
+                p.id for p in Product.browse(product_ids) if p.consumable}
             for (location_id, product), quantity in quantities.items():
-                if quantity:
+                if quantity and product not in consumables:
                     empty.discard(location_id)
             for sub_ids in grouped_slice(list(empty)):
                 sub_ids = list(sub_ids)
                 moves = Move.search([
-                        ('state', 'not in', ['done', 'cancel']),
+                        ('state', 'not in', ['done', 'cancelled']),
                         ['OR',
                             ('from_location', 'in', sub_ids),
                             ('to_location', 'in', sub_ids),
@@ -285,13 +343,9 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
     def default_type():
         return 'storage'
 
-    @classmethod
-    def check_xml_record(cls, records, values):
-        return True
-
     def get_warehouse(self, name):
         # Order by descending left to get the first one in the tree
-        with Transaction().set_context(active_test=False):
+        with inactive_records():
             locations = self.search([
                     ('parent', 'parent_of', [self.id]),
                     ('type', '=', 'warehouse'),
@@ -317,6 +371,17 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             cls._default_warehouse_cache.set(None, warehouse)
         return warehouse
 
+    @property
+    def lost_found_used(self):
+        if self.warehouse:
+            return self.warehouse.lost_found_location
+
+    def get_rec_name(self, name):
+        if self.code:
+            return f'[{self.code}] {self.name}'
+        else:
+            return self.name
+
     @classmethod
     def search_rec_name(cls, name, clause):
         if clause[1].startswith('!') or clause[1].startswith('not '):
@@ -328,6 +393,20 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             ('code',) + tuple(clause[1:]),
             ]
 
+    @classmethod
+    def _get_quantity_grouping(cls):
+        context = Transaction().context
+        grouping, grouping_filter, key = (), (), []
+        if context.get('product') is not None:
+            grouping = ('product',)
+            grouping_filter = ([context['product']],)
+            key = (context['product'],)
+        elif context.get('product_template') is not None:
+            grouping = ('product.template',)
+            grouping_filter = ([context['product_template']],)
+            key = (context['product_template'],)
+        return grouping, grouping_filter, key
+
     @classmethod
     def get_quantity(cls, locations, name):
         pool = Pool()
@@ -339,12 +418,9 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             return (trans_context.get(name) is not None
                 and isinstance(trans_context[name], int))
 
-        if not any(map(valid_context, ['product', 'product_template'])):
-            return {l.id: None for l in locations}
-
         context = {}
         if (name == 'quantity'
-                and (trans_context.get('stock_date_end', datetime.date.max)
+                and ((trans_context.get('stock_date_end') or datetime.date.max)
                     > Date_.today())):
             context['stock_date_end'] = Date_.today()
 
@@ -353,14 +429,10 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             if not trans_context.get('stock_date_end'):
                 context['stock_date_end'] = datetime.date.max
 
-        if trans_context.get('product') is not None:
-            grouping = ('product',)
-            grouping_filter = ([trans_context['product']],)
-            key = trans_context['product']
-        else:
-            grouping = ('product.template',)
-            grouping_filter = ([trans_context['product_template']],)
-            key = trans_context['product_template']
+        grouping, grouping_filter, key = cls._get_quantity_grouping()
+        if not grouping:
+            return {loc.id: None for loc in locations}
+
         pbl = {}
         for sub_locations in grouped_slice(locations):
             location_ids = [l.id for l in sub_locations]
@@ -371,7 +443,7 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
                         grouping_filter=grouping_filter,
                         with_childs=trans_context.get('with_childs', True)))
 
-        return dict((loc.id, pbl.get((loc.id, key), 0)) for loc in locations)
+        return dict((loc.id, pbl.get((loc.id,) + key, 0)) for loc in locations)
 
     @classmethod
     def search_quantity(cls, name, domain):
@@ -393,6 +465,24 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
                 ids.append(location.id)
         return [('id', 'in', ids)]
 
+    @classmethod
+    def get_quantity_uom(cls, locations, name):
+        pool = Pool()
+        Product = pool.get('product.product')
+        Template = pool.get('product.template')
+        context = Transaction().context
+        value = None
+        uom = None
+        if context.get('product') is not None:
+            product = Product(context['product'])
+            uom = product.default_uom
+        elif context.get('product_template') is not None:
+            template = Template(context['product_template'])
+            uom = template.default_uom
+        if uom:
+            value = uom.id
+        return {l.id: value for l in locations}
+
     @classmethod
     def get_cost_value(cls, locations, name):
         pool = Pool()
@@ -415,7 +505,7 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
                 return Template(trans_context['product_template'])
 
         context = {}
-        if 'stock_date_end' in trans_context:
+        if trans_context.get('stock_date_end') is not None:
             # Use the last cost_price of the day
             context['_datetime'] = datetime.datetime.combine(
                 trans_context['stock_date_end'], datetime.time.max)
@@ -428,7 +518,7 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
         # The template may have more than one product
         if cost_price is not None:
             for location in locations:
-                cost_values[location.id] = (
+                cost_values[location.id] = round_price(
                     Decimal(str(location.quantity)) * cost_price)
         return cost_values
 
@@ -455,48 +545,35 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
         cls.save(to_save)
 
     @classmethod
-    def create(cls, vlist):
-        locations = super(Location, cls).create(vlist)
-        cls._set_warehouse_parent(locations)
+    def on_modification(cls, mode, locations, field_names=None):
+        super().on_modification(mode, locations, field_names=field_names)
+        if mode in {'create', 'write'}:
+            cls._set_warehouse_parent(locations)
         cls._default_warehouse_cache.clear()
-        return locations
 
     @classmethod
-    def write(cls, *args):
-        super(Location, cls).write(*args)
-        locations = sum(args[::2], [])
-        cls._set_warehouse_parent(locations)
-        cls._default_warehouse_cache.clear()
-
-        ids = [l.id for l in locations]
-        warehouses = cls.search([
-                ('type', '=', 'warehouse'),
-                ['OR',
-                    ('storage_location', 'in', ids),
-                    ('input_location', 'in', ids),
-                    ('output_location', 'in', ids),
-                    ]])
-
-        fields = ('storage_location', 'input_location', 'output_location')
-        wh2childs = {}
-        for warehouse in warehouses:
-            in_out_sto = (getattr(warehouse, f).id for f in fields)
-            for location in locations:
-                if location.id not in in_out_sto:
-                    continue
-                childs = wh2childs.setdefault(warehouse.id, cls.search([
-                            ('parent', 'child_of', warehouse.id),
-                            ]))
-                if location not in childs:
-                    raise LocationValidationError(
-                        gettext('stock.msg_location_child_of_warehouse',
-                            location=location.rec_name,
-                            warehouse=warehouse.rec_name))
+    def check_modification(cls, mode, locations, values=None, external=False):
+        super().check_modification(
+            mode, locations, values=values, external=external)
+        if mode == 'write' and 'parent' in values:
+            warehouses = cls.search([
+                    ('type', '=', 'warehouse'),
+                    ])
+            cls.validate_fields(
+                warehouses,
+                {'storage_location', 'input_location', 'output_location'})
 
     @classmethod
-    def delete(cls, *args):
-        super().delete(*args)
-        cls._default_warehouse_cache.clear()
+    def delete(cls, locations):
+        # Delete also required children as CASCADING is done separately
+        extra_locations = []
+        for location in locations:
+            extra_locations.extend(filter(None, [
+                        location.input_location,
+                        location.output_location,
+                        location.storage_location,
+                        ]))
+        super().delete(locations + extra_locations)
 
     @classmethod
     def copy(cls, locations, default=None):
@@ -516,7 +593,7 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
                 wh_default['storage_location'] = None
                 wh_default['childs'] = None
 
-                new_location, = super(Location, cls).copy([location],
+                new_location, = super().copy([location],
                     default=wh_default)
 
                 with Transaction().set_context(
@@ -532,7 +609,7 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
                         'type': 'warehouse',
                         })
             else:
-                new_location, = super(Location, cls).copy([location],
+                new_location, = super().copy([location],
                     default=default)
                 warehouse_locations = Transaction().context.get(
                     'cp_warehouse_locations') or {}
@@ -548,88 +625,26 @@ class Location(DeactivableMixin, tree(), ModelSQL, ModelView):
             res.append(new_location)
         return res
 
-
-supplier_location = fields.Many2One(
-    'stock.location', "Supplier Location", domain=[('type', '=', 'supplier')],
-    help="The default source location for stock received from the party.")
-customer_location = fields.Many2One(
-    'stock.location', "Customer Location", domain=[('type', '=', 'customer')],
-    help="The default destination location for stock sent to the party.")
-
-
-class Party(metaclass=PoolMeta):
-    __name__ = 'party.party'
-    supplier_location = fields.MultiValue(supplier_location)
-    customer_location = fields.MultiValue(customer_location)
-    locations = fields.One2Many(
-        'party.party.location', 'party', "Locations")
-
     @classmethod
-    def multivalue_model(cls, field):
-        pool = Pool()
-        if field in {'supplier_location', 'customer_location'}:
-            return pool.get('party.party.location')
-        return super(Party, cls).multivalue_model(field)
-
-    @classmethod
-    def default_supplier_location(cls, **pattern):
-        return cls.multivalue_model(
-            'supplier_location').default_supplier_location()
-
-    @classmethod
-    def default_customer_location(cls, **pattern):
-        return cls.multivalue_model(
-            'customer_location').default_customer_location()
-
-
-class PartyLocation(ModelSQL, ValueMixin):
-    "Party Location"
-    __name__ = 'party.party.location'
-    party = fields.Many2One(
-        'party.party', "Party", ondelete='CASCADE', select=True)
-    supplier_location = supplier_location
-    customer_location = customer_location
-
-    @classmethod
-    def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        exist = TableHandler.table_exist(cls._table)
-
-        super(PartyLocation, cls).__register__(module_name)
-
-        if not exist:
-            cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.extend(['supplier_location', 'customer_location'])
-        value_names.extend(['supplier_location', 'customer_location'])
-        migrate_property(
-            'party.party', field_names, cls, value_names,
-            parent='party', fields=fields)
-
-    @classmethod
-    def default_supplier_location(cls):
-        pool = Pool()
-        ModelData = pool.get('ir.model.data')
-        try:
-            return ModelData.get_id('stock', 'location_supplier')
-        except KeyError:
-            return None
-
-    @classmethod
-    def default_customer_location(cls):
-        pool = Pool()
-        ModelData = pool.get('ir.model.data')
-        try:
-            return ModelData.get_id('stock', 'location_customer')
-        except KeyError:
-            return None
+    def view_attributes(cls):
+        storage_types = Eval('type').in_(['storage', 'warehouse', 'view'])
+        return super().view_attributes() + [
+            ('/tree/field[@name="quantity"]',
+                'visual', If(
+                    storage_types & (Eval('quantity', 0) < 0), 'danger', ''),
+                ['type']),
+            ('/tree/field[@name="forecast_quantity"]',
+                'visual', If(
+                    storage_types & (Eval('forecast_quantity', 0) < 0),
+                    'warning', ''),
+                ['type']),
+            ]
 
 
 class ProductsByLocationsContext(ModelView):
-    'Products by Locations'
     __name__ = 'stock.products_by_locations.context'
+
+    company = fields.Many2One('company.company', "Company", required=True)
     forecast_date = fields.Date(
         'At Date',
         help="The date for which the stock quantity is calculated.\n"
@@ -638,6 +653,10 @@ class ProductsByLocationsContext(ModelView):
     stock_date_end = fields.Function(fields.Date('At Date'),
         'on_change_with_stock_date_end')
 
+    @classmethod
+    def default_company(cls):
+        return Transaction().context.get('company')
+
     @staticmethod
     def default_forecast_date():
         Date_ = Pool().get('ir.date')
@@ -650,21 +669,88 @@ class ProductsByLocationsContext(ModelView):
         return self.forecast_date
 
 
+class ProductsByLocations(DeactivableMixin, ModelSQL, ModelView):
+    __name__ = 'stock.products_by_locations'
+
+    product = fields.Many2One('product.product', "Product")
+    quantity = fields.Function(
+        fields.Float("Quantity", digits='default_uom'),
+        'get_product', searcher='search_product')
+    forecast_quantity = fields.Function(
+        fields.Float("Forecast Quantity", digits='default_uom'),
+        'get_product', searcher='search_product')
+    default_uom = fields.Function(
+        fields.Many2One(
+            'product.uom', "Default UoM",
+            help="The default Unit of Measure."),
+        'get_product', searcher='search_product')
+    cost_value = fields.Function(
+        fields.Numeric("Cost Value"), 'get_product')
+    consumable = fields.Function(
+        fields.Boolean("Consumable"), 'get_product',
+        searcher='search_product')
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('product', 'ASC'))
+
+    @classmethod
+    def table_query(cls):
+        pool = Pool()
+        Product = pool.get('product.product')
+        product = Product.__table__()
+        columns = []
+        for fname, field in cls._fields.items():
+            if not hasattr(field, 'set'):
+                if (isinstance(field, fields.Many2One)
+                        and field.get_target() == Product):
+                    column = Column(product, 'id')
+                else:
+                    column = Column(product, fname)
+                columns.append(column.as_(fname))
+        return product.select(*columns)
+
+    def get_rec_name(self, name):
+        return self.product.rec_name
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('product.rec_name',) + tuple(clause[1:])]
+
+    def get_product(self, name):
+        value = getattr(self.product, name)
+        if isinstance(value, Model):
+            value = value.id
+        return value
+
+    @classmethod
+    def search_product(cls, name, clause):
+        nested = clause[0][len(name):]
+        return [('product.' + name + nested, *clause[1:])]
+
+
 class LocationLeadTime(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
-    'Location Lead Time'
     __name__ = 'stock.location.lead_time'
 
     warehouse_from = fields.Many2One('stock.location', 'Warehouse From',
         ondelete='CASCADE',
         domain=[
             ('type', '=', 'warehouse'),
+            ('id', '!=', Eval('warehouse_to', -1)),
             ])
     warehouse_to = fields.Many2One('stock.location', 'Warehouse To',
         ondelete='CASCADE',
         domain=[
             ('type', '=', 'warehouse'),
+            ('id', '!=', Eval('warehouse_from', -1)),
             ])
-    lead_time = fields.TimeDelta('Lead Time',
+    lead_time = fields.TimeDelta(
+        "Lead Time",
+        domain=['OR',
+            ('lead_time', '=', None),
+            ('lead_time', '>=', TimeDelta()),
+            ],
         help="The time it takes to move stock between the warehouses.")
 
     @classmethod
